@@ -28,9 +28,9 @@ typedef struct connection_tag
     unsigned long id;
 
     time_t con_time;
+    time_t discon_time;
     uint64_t sent_bytes;
 
-
     int sock;
     int serversock;
     int error;
@@ -38,21 +38,16 @@ typedef struct connection_tag
     char *ip;
     char *host;
 
-    /* For 'fake' connections */
-    int event_number;
-    void *event;
 } connection_t;
 
 void connection_initialize(void);
 void connection_shutdown(void);
 void connection_accept_loop(void);
 void connection_close(connection_t *con);
-connection_t *create_connection(sock_t sock, sock_t serversock, char *ip);
-int connection_complete_source (struct source_tag *source);
+connection_t *connection_create (sock_t sock, sock_t serversock, char *ip);
+int connection_complete_source (struct source_tag *source, int response);
 
-void connection_inject_event(int eventnum, void *event_data);
-
-int connection_check_source_pass(http_parser_t *parser, char *mount);
+int connection_check_source_pass(http_parser_t *parser, const char *mount);
 int connection_check_relay_pass(http_parser_t *parser);
 int connection_check_admin_pass(http_parser_t *parser);
 
