@@ -24,6 +24,7 @@ typedef struct _relay_server {
     char *localmount;
     struct source_tag *source;
     int mp3metadata;
+    int on_demand;
     int running;
     int cleanup;
     thread_type *thread;
@@ -33,7 +34,9 @@ typedef struct _relay_server {
 
 void slave_initialize(void);
 void slave_shutdown(void);
-void slave_recheck (void);
+void slave_recheck_mounts (void);
+void slave_rebuild_mounts (void);
+void slave_rescan (void);
 relay_server *relay_free (relay_server *relay);
 
 #endif  /* __SLAVE_H__ */
